#!/usr/bin/env python3
# Copyright (c) 2004-present Facebook All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

ENUM_DIRNAME = "enum"
INPUT_DIRNAME = "input"

SPACES = " " * 4
